import json
import jsonpickle
import os
from pprint import pformat
import re

from sectile import Sectile


class MissingValue(Exception):
    pass


class ContextMixin:
    def get_context_data(self):
        context = {}
        context['site'] = self.flourish.site_config
        context['global'] = self.flourish.global_context(self.flourish)
        context['tokens'] = self.tokens
        if self.context:
            context.update(**self.context)
        return context


class PathMixin:
    PATH_SEGMENTS = r'(#\w+)'
    STRIP_PATH = r'^[\w_-]+'

    def get_current_path(self, tokens):
        self.current_path = self.resolve(**tokens)
        return self.current_path

    def resolve(self, **kwargs):
        resolved = ''
        # FIXME use self.arguments
        for segment in re.split(self.PATH_SEGMENTS, self.path):
            if segment.startswith('#'):
                key = segment[1:]
                if key in kwargs:
                    if kwargs[key] is None:
                        raise RuntimeError
                    if key == 'month' or key == 'day':
                        resolved = '%s%02d' % (resolved, int(kwargs[key]))
                    else:
                        resolved = '%s%s' % (resolved, kwargs[key])
                else:
                    raise KeyError
            else:
                resolved = resolved + segment
        return resolved

    def can_generate(self, path):
        subpath = None
        if path.endswith('?'):
            subpath = path[:-1]
        generates = []

        # see if the path can be rejected before continuing to the
        # expensive check-all-possible-matches, by checking if each
        # non-token part of the requested path is found in our path
        check_path = path
        segments = re.split(self.PATH_SEGMENTS, self.path)
        while len(segments) > 1:
            if check_path == '?':
                break
            pos = check_path.find(segments[0])
            if pos < 0:
                return []
            check_path = check_path[(pos + len(segments[0])):]
            stripped = re.split(self.STRIP_PATH, check_path)
            if segments[1] == '#slug' and check_path.endswith('/'):
                path = path + 'index'
            elif len(stripped) > 1:
                check_path = stripped[1]
            segments.pop(0)
            segments.pop(0)
        if segments[0]:
            if check_path.endswith('?'):
                if segments[0].find(check_path[:-1]) < 0:
                    return []
            elif not check_path.endswith(segments[0]):
                return []

        for _filter in self.all_valid_filters():
            filter_path = self.resolve(**_filter)
            if path == filter_path:
                generates.append(_filter)
            elif subpath and filter_path.startswith(subpath):
                generates.append(_filter)
        return generates

    def all_valid_filters(self):
        valid_filters = []
        args = self.arguments

        if len(args) == 0:
            valid_filters.append({})
        else:
            filters = self.flourish.get_valid_filters_for_tokens(
                args,
                self.get_filtered_sources(),
            )
            for _filter in filters:
                valid_filters.append(_filter)
        return valid_filters

    @property
    def arguments(self):
        arguments = []
        for segment in re.split(self.PATH_SEGMENTS, self.path):
            if segment.startswith('#'):
                arguments.append(segment[1:])
        return arguments

    def get_context_data(self):
        context = super().get_context_data()
        context['current_path'] = self.current_path
        return context


class SourcesMixin:
    order_by = None
    sources_exclude = None
    sources_filter = None
    limit = None

    def get_objects(self, tokens):
        sources = self.get_filtered_sources().filter(**tokens)
        ordering = self.get_order_by()
        if ordering is not None:
            sources = sources.order_by(self.order_by)
        if self.limit is not None:
            self.source_objects = sources[0:self.limit]
        else:
            self.source_objects = sources
        return self.source_objects

    def get_filtered_sources(self):
        sources = self.flourish.sources
        if self.sources_filter is not None:
            sources = sources.filter(**self.sources_filter)
        if self.sources_exclude is not None:
            sources = sources.exclude(**self.sources_exclude)
        return sources

    def get_order_by(self):
        return self.order_by

    def get_context_data(self):
        context = super().get_context_data()
        context['sources'] = self.source_objects
        return context


class GeneratorMixin:
    class DoNotGenerate(Exception):
        pass

    def generate(self, report=False, tokens=None):
        self.report = report
        if not tokens:
            tokens = self.get_path_tokens()
        for tokenset in tokens:
            self.tokens = tokenset
            try:
                self.generate_path(tokenset)
            except self.DoNotGenerate:
                pass

    def get_path_tokens(self):
        return self.flourish.all_valid_filters_for_path(self.name)

    def generate_path(self, tokens):
        self.get_current_path(tokens)
        self.get_objects(tokens)
        self.output_to_file()

    def get_blueprint_context_as_strings(self):
        # jump through hoops to get context "pretty" printed without
        # the ridiculous indentation that pprint generates
        context = self.get_context_data()
        global_context = context['global']
        del context['global']

        obj = json.loads(
                jsonpickle.encode(context, unpicklable=False)
            )
        # do not need to see all this data in the blueprint page
        if 'page' in obj:
            del obj['page']['_parent']['_cache']
        context_string = json.dumps(obj, indent=2)

        global_context_string = json.dumps(
            json.loads(
                jsonpickle.encode(global_context, unpicklable=False)
            ),
            indent=2,
        )
        return context_string, global_context_string

    def get_blueprint(self, tokens):
        for tokenset in tokens:
            self.tokens = tokenset
            self.get_current_path(tokenset)
            self.get_objects(tokenset)
            context = self.get_context_data()
            name = self.get_template_name()

            page_context_string, global_context_string \
                = self.get_blueprint_context_as_strings()

            # FIXME no pages in context on index page
            blueprint = {
                'path': self.current_path,
                'context': context,
                'debug_page_context': page_context_string,
                'debug_global_context': global_context_string,
                'template_name': name,
            }
            if self.flourish.using_sectile:
                sectile = Sectile(fragments=self.flourish.fragments_dir)
                dimensions = {}
                for dimension in self.flourish.jinja.loader.dimensions():
                    if dimension in context:
                        dimensions[dimension] = context[dimension]
                (
                    blueprint['template'],
                    blueprint['sectile_fragments'] 
                ) = sectile.generate(
                        self.current_path,
                        name,
                        **dimensions
                    )
                blueprint['sectile_dimension_possibilities'] \
                    = sectile.get_dimension_possibilities(
                            self.current_path,
                            **dimensions
                        )
                # expand unspecified dimensions for debugging
                for dimension in self.flourish.jinja.loader.dimensions():
                    if dimension not in dimensions:
                        dimensions[dimension] = 'all'
                blueprint['sectile_dimensions'] = dimensions
                blueprint['sectile_fragment_paths'] \
                    = sectile.get_dimension_paths(
                            self.current_path,
                            **dimensions,
                        )
            else:
                filename = os.path.join(self.flourish.templates_dir, name)
                with open(filename, 'r') as handle:
                    blueprint['template'] = handle.read()
            return blueprint


class TemplateMixin:
    def render_output(self):
        context = self.get_context_data()
        template = self.get_template(context)
        return self.render_template(template, context)

    def get_template(self, context):
        name = self.get_template_name()
        if name is None:
            raise MissingValue
        if self.flourish.using_sectile:
            # Sectile templates need to be constructed rather than just found
            # on the filesystem. Because the entrypoint to the jinja loader
            # doesn't allow any extra context than the name of the template to
            # be passed, we pre-construct and cache a template and then use
            # that cached value as the template's "name".
            dimensions = {}
            for dimension in self.flourish.jinja.loader.dimensions():
                if dimension in context:
                    dimensions[dimension] = context[dimension]
            digest = self.flourish.jinja.loader.prepare_template(
                self.current_path,
                name,
                **dimensions,
            )
            return self.flourish.jinja.get_template(digest)
        else:
            return self.flourish.jinja.get_template(name)

    def get_template_name(self):
        return self.template_name

    def render_template(self, template, context_data):
        return template.render(context_data)


class PageTemplateMixin:
    template_name = 'page.html'

    def get_template_name(self):
        if self.source_objects:
            if 'template' in self.source_objects[0]:
                return self.source_objects[0]['template']
            if 'page_type' in self.source_objects[0]:
                return '%s.html' % self.source_objects[0]['page_type']
        return super().get_template_name()



from datetime import date

from flourish.generators.base import IndexGenerator


class CalendarGenerator(IndexGenerator):
    order_by = 'published'


class CalendarDayGenerator(CalendarGenerator):
    template_name = 'calendar_day.html'

    def get_context_data(self):
        _context = super().get_context_data()
        _context['day'] = date(
            day=int(self.tokens['day']),
            month=int(self.tokens['month']),
            year=int(self.tokens['year']),
        )
        return _context


class CalendarMonthGenerator(CalendarGenerator):
    template_name = 'calendar_month.html'

    def get_context_data(self):
        _context = super().get_context_data()
        _context['month'] = date(
            day=int(1),
            month=int(self.tokens['month']),
            year=int(self.tokens['year']),
        )
        dates = self.source_objects.publication_dates
        _context['publication_dates'] = dates[0]['months'][0]['days']
        return _context


class CalendarYearGenerator(CalendarGenerator):
    template_name = 'calendar_year.html'

    def get_context_data(self):
        _context = super().get_context_data()
        _context['year'] = date(
            day=1,
            month=1,
            year=int(self.tokens['year']),
        )
        dates = self.source_objects.publication_dates
        _context['publication_dates'] = dates[0]['months']
        return _context



import codecs
import csv
from datetime import datetime, timezone
from glob import glob
import json
import os
import re
import warnings

import markdown2
import toml


class SourceBase:
    @property
    def slug(self):
        return self._slug

    @property
    def path(self):
        # FIXME when _source_url is not set
        _path = self._parent._source_path
        _filter = {}
        for _arg in _path.arguments:
            try:
                _value = getattr(self, _arg)
                if type(_value) == list:
                    _filter[_arg] = _value[0]
                else:
                    _filter[_arg] = _value
            except AttributeError:
                warnings.warn(
                    'cannot create URL for "%s": no value for "%s"' % (
                        self._slug, _arg)
                )
                return None
        _resolved = _path.resolve(**_filter)
        if _resolved.endswith('/index'):
            _resolved = _resolved[:-5]
        return _resolved

    @property
    def absolute_url(self):
        return '%s%s' % (self._parent.site_config['base_url'], self.path)

    @property
    def timestamp(self):
        return self._timestamp

    def related(self, key):
        try:
            _filter = {}
            _filter[key] = self._config[key]
            return self._parent.sources.filter(
                    **_filter
                ).exclude(
                    slug=self.slug
                )
        except KeyError:
            return []

    def _add_html_attachments(self):
        pattern = '%s/%s.*.html' % (self._parent.source_dir, self.slug)
        for attachment in glob(pattern):
            key = attachment.split('.')[-2]
            with codecs.open(attachment, encoding='utf-8') as content:
                if key in self._config and len(self._config[key]):
                    warnings.warn(
                        '"%s" in %s overriden by HTML attachment.' % (
                            key, self.slug))
                self._config[key] = content.read()

    def _add_markdown_attachments(self):
        pattern = '%s/%s.*.markdown' % (self._parent.source_dir, self.slug)
        for attachment in glob(pattern):
            key = attachment.split('.')[-2] + '_markdown'
            with codecs.open(attachment, encoding='utf-8') as content:
                if key in self._config and len(self._config[key]):
                    warnings.warn(
                        '"%s" in %s overriden by Markdown attachment.' % (
                            key, self.slug))
                self._config[key] = content.read()

    def _convert_markdown(self):
        add = {}
        for key in self._config:
            if key[-9:] == '_markdown':
                dest = key[:-9]
                add[dest] = markdown2.markdown(self._config[key])
                if dest in self._config and len(self._config[dest]):
                    warnings.warn(
                        '"%s" in %s overriden by Markdown conversion.' % (
                            dest, self.slug))
        self._config.update(add)

    def __getattr__(self, key):
        if key == 'slug':
            return self.slug
        if key in self._config:
            return self._config[key]

        # foreign key lookup
        _fkey = '%s_fkey' % key
        if _fkey in self._config:
            return self._parent.get(self._config[_fkey])

        # foreign key reverse lookup
        if key.endswith('_set'):
            _fkey = '%s_fkey' % key[:-4]
            _filter = {_fkey: self.slug}
            return self._parent.sources.filter(**_filter)

        if key in ['body', 'title']:
            # these are required for atom feeds, so if they've not been
            # set in the page's configuration, return them as empty
            return ''
        raise AttributeError

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __iter__(self):
        _keys = list(self._config)
        _keys.append('slug')
        return iter(_keys)

    class DoesNotExist(Exception):
        pass


class SourceFile(SourceBase):
    def __init__(self, parent, filename):
        slug, _ = os.path.splitext(filename)
        self._source = filename
        self._slug = slug
        self._parent = parent
        self._config = self._read_configuration(filename)
        self._timestamp = os.stat(
                os.path.join(self._parent.source_dir, filename)
            ).st_mtime
        self._add_markdown_attachments()
        self._convert_markdown()
        self._add_html_attachments()

    def _read_configuration(self, filename):
        toml_file = '%s/%s' % (self._parent.source_dir, filename)
        with codecs.open(toml_file, encoding='utf-8') as configuration:
            return toml.loads(configuration.read())

    def __repr__(self):
        return '<flourish.SourceFile object (%s)>' % self._source


class MarkdownSourceFile(SourceFile):
    def _read_configuration(self, filename):
        markdown_file = '%s/%s' % (self._parent.source_dir, filename)
        with codecs.open(markdown_file, encoding='utf-8') as configuration:
            content = configuration.read()
        if content.startswith('-') or content.startswith('`'):
            _delim_char = content[0:1]
            FM_SPLIT = re.compile('^%s{3}$' % _delim_char, re.MULTILINE)
            try:
                _, _frontmatter, _body = FM_SPLIT.split(content, 2)
                config = toml.loads(_frontmatter)
                config['body_markdown'] = _body
            except ValueError:
                raise RuntimeError(
                    '"%s" has no end marker for the frontmatter' % filename
                )
        else:
            config = {}
            config['body_markdown'] = content
        return config

    def __repr__(self):
        return '<flourish.MarkdownSourceFile object (%s)>' % self._source


class TomlSourceFile(SourceFile):
    def __repr__(self):
        return '<flourish.TomlSourceFile object (%s)>' % self._source


class JsonSourceFile(SourceFile):
    ISO8601 = re.compile(r'^\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}Z$')

    def _read_configuration(self, filename):
        _json_file = '%s/%s' % (self._parent.source_dir, filename)
        with codecs.open(_json_file, encoding='utf-8') as _configuration:
            _config = json.loads(_configuration.read())

        for _key, _value in _config.items():
            if type(_value) == str and self.ISO8601.match(_value):
                _config[_key] = datetime.strptime(
                        _value, "%Y-%m-%dT%H:%M:%SZ"
                    ).replace(tzinfo=timezone.utc)
        return _config

    def __repr__(self):
        return '<flourish.JsonSourceFile object (%s)>' % self._source


class CsvRowSource(SourceBase):
    ISO8601 = re.compile(r'^\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}Z$')

    def __init__(self, parent, filename, index, row):
        self._parent = parent
        self._source = filename
        self._index = index
        self._config = row
        if row['slug'].startswith('/'):
            self._slug = row['slug'][1:]
        else:
            self._slug = row['slug']
        if self._slug.endswith('/'):
            self._slug += 'index'

        _add = {}
        for _key, _value in row.items():
            # convert timestamps
            if type(_value) == str and self.ISO8601.match(_value):
                row[_key] = datetime.strptime(
                        _value, "%Y-%m-%dT%H:%M:%SZ"
                    ).replace(tzinfo=timezone.utc)
            # split string arrays
            if _key.endswith('[]'):
                _akey = _key[:-2]
                if _value.find(',') > -1:
                    _add[_akey] = _value.split(',')
                else:
                    _add[_akey] = _value.split(':')
        self._config.update(_add)

        # clean up unwanted keys from the original CSV
        del(row['slug'])
        for _key in _add:
            del(row['%s[]' % _key])

        self._add_markdown_attachments()
        self._convert_markdown()
        self._add_html_attachments()

    def __repr__(self):
        return '<flourish.CsvRowSource object (%s, row %d)>' % (
                self._source,
                self._index,
            )


class MultipleSourcesFile:
    def __init__(self, parent, filename):
        self._parent = parent
        self._sources = self._read_file(filename)

    def get_sources(self):
        return self._sources


class CsvSourceFile(MultipleSourcesFile):
    VALID_SLUG = re.compile(r'^/?[a-zA-Z0-9][a-zA-Z0-9_/-]*$')

    def _read_file(self, filename):
        _sources = []
        _source = '%s/%s' % (self._parent.source_dir, filename)
        with open(_source) as handle:
            reader = csv.DictReader(handle)
            for line, row in enumerate(reader):
                if 'slug' not in row:
                    warnings.warn('"%s" has no column "slug"' % filename)
                    break
                else:
                    if not self.VALID_SLUG.match(row['slug']):
                        warnings.warn(
                            '"%s" row %d, has an invalid slug "%s"' % (
                                filename,
                                line,
                                row['slug'],
                            )
                        )
                        continue
                    _sources.append(
                        CsvRowSource(self._parent, filename, line, row)
                    )
        return _sources



from datetime import date


def ordinal(value):
    if value in (11, 12, 13):
        return 'th'
    last = value % 10
    if last == 1:
        return 'st'
    if last == 2:
        return 'nd'
    if last == 3:
        return 'rd'
    return 'th'


def month_name(value):
    return date(1970, value, 1).strftime('%B')



import os


class NotPath(Exception):
    pass


class TrieNode:
    def __init__(self, path):
        self.path = path
        self.children = dict()
        self.published = False


class DirTrie:
    """
    A DirTrie is a trie-like structure, splitting paths per directory
    (whereas a class trie builds a character-based tree).

    Its primary purpose is to take a list of generated paths, and collapse
    that list to a smaller number for use in CloudFront invalidation requests.
    Republishing an entire site with >1000 paths (eg the footer changes,
    so the entire site is rebuilt) edges into paid-for invalidation costs.
    Being able to collapse a list into a smaller list of invalidation paths,
    even at the expense of expunging valid paths from the edge cache works
    out cheaper for small traffic sites.
    """
    def __init__(self):
        self.root = TrieNode('/')

    def split_paths(self, path):
        """
        Turn '/a/b/c' into [/, /a/, /a/b/, /a/b/c] for ease of use.
        """
        if not path.startswith('/'):
            raise NotPath
        paths = [path,]
        while path != '/':
            head, tail = os.path.split(path)
            if tail != '':
                paths.insert(0, os.path.join(head, ''))
            path = head
        return paths

    def insert(self, path):
        """
        Insert a new published path into the trie.
        """
        node = self.root
        paths = self.split_paths(path)
        for path in paths:
            if path == '/':
                continue
            if path not in node.children:
                node.children[path] = TrieNode(path=path)
            node = node.children[path]
        node.published = True

    def size(self, node=None):
        """
        Count the number of nodes below the specified node
        (default: root node).
        """
        if not node:
            node = self.root
        count = 1
        for child in node.children:
            count += self.size(node.children[child])
        return count

    def published_count(self, node=None):
        """
        Count the number of nodes that represent published paths below the
        specied node (default: root).
        """
        if not node:
            node = self.root
        count = 0
        if node.published:
            count = 1
        for child in node.children:
            count += self.published_count(node.children[child])
        return count

    def published_paths(self, node=None):
        """
        Return the list of nodes that represent published paths below
        the specified node (default: root node).
        """
        paths = []
        if node is None:
            node = self.root
        if node.published:
            paths.append(node.path)
        for child in node.children:
            paths.extend(self.published_paths(node.children[child]))
        return paths

    def find(self, path):
        """
        Find the node for a given path, if it exists.
        """
        node = self.root
        paths = self.split_paths(path)
        for path in paths:
            if path == '/':
                continue
            if path not in node.children:
                return None
            node = node.children[path]
        return node

    def collapse(self, max, node=None):
        """
        Reduce the number of published paths to `max` or below for CloudFront
        invalidation. Packs in the subpaths with the smallest number of
        changes first, then the bigger ones. Anything with too many paths is
        returned as '/a/b/*', a wildcard invalidation which can invalidate
        many paths, but is only billed as one path.
        """
        if node is None:
            node = self.root

        if len(self.published_paths(node)) <= max:
            return self.published_paths(node)

        paths = []
        if node.children:
            if len(node.children) > max:
                return ['%s*' % node.path]

            children = [
                x for x in
                    sorted(
                        node.children,
                        key=lambda c: self.published_count(node.children[c]),
                    )
            ]
            remains = max - len(children)
            if node.published:
                remains -= 1
            for child in children:
                new_paths = self.collapse(remains + 1, node.children[child])
                paths.extend(new_paths)
                remains -= len(new_paths) - 1

        if node.published:
            paths.append(node.path)

        if max - len(paths) < 0:
            paths = ['%s*' % node.path]

        return paths

    def print_trie(self, node=None, depth=0):
        """
        Print the trie structure, for debugging.
        """
        if node is None:
            node = self.root
        pub_icon = '--'
        if node.published:
            pub_icon = '**'
        print('  '*depth, pub_icon, node.path)
        for child in node.children:
            self.print_trie(node.children[child], depth+1)



import csv
from datetime import datetime

from flourish.generators.base import BaseGenerator
from flourish.generators.mixins import SourcesMixin


class CSVGenerator(SourcesMixin, BaseGenerator):
    file_extension = '.csv'
    fields = ['title', 'published']

    def output_to_file(self):
        filename = self.get_output_filename()
        with open(filename, 'w', newline='') as handle:
            output = csv.DictWriter(
                handle,
                fieldnames = self.get_fields(),
                quoting=csv.QUOTE_MINIMAL,
            )
            output.writeheader()

            for object in self.source_objects:
                output.writerow(self.get_row(object))

    def get_fields(self):
        return self.fields

    def get_row(self, object):
        row = {}
        for field in self.get_fields():
            row.update({field: self.get_field_value(object, field)})
        return row

    def get_field_value(self, object, field):
        value = getattr(object, field, None)
        if type(value) == list:
            if type(value[0]) == str:
                return ':'.join(sorted(value))
        elif type(value) == datetime:
            return value.isoformat().replace('+00:00', 'Z')
        return value



from setuptools import find_packages, setup

# read in __version__
with open('flourish/version.py') as version_file:
    exec(version_file.read())

setup(
    name='flourish',
    version=__version__,    # noqa: F821

    description='Static website generator',
    url='https://github.com/norm/flourish',

    license='MIT',
    author='Mark Norman Francis',
    author_email='norm@201created.com',

    packages=find_packages(exclude=['tests']),
    entry_points={
        'console_scripts': ['flourish=flourish.command_line:main'],
    },

    package_data={
        '': ['*.html'],
    },

    install_requires=[
        'boto3',
        'Flask',
        'Jinja2>=2.9',
        'markdown2',
        'feedgen',
        'toml',
        'libsass',
        'sectile>=0.3.5',
        'jsonpickle',
        'importlib_resources; python_version <= "3.9"',
    ],
    python_requires='>=3.6',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',

        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
)



import os

from .mixins import (
    PathMixin,
    SourcesMixin,
    GeneratorMixin,
    ContextMixin,
    TemplateMixin,
    PageTemplateMixin,
)
from ..paginator import Paginator


class BaseGenerator(ContextMixin, PathMixin, GeneratorMixin, TemplateMixin):
    """
    The BaseGenerator takes care of the interface with Flourish.
    """
    def __init__(self, path, name, context={}):
        self.path = path
        self.name = name
        self.context = context

    def setup(self, flourish):
        self.flourish = flourish

    def get_context_data(self):
        context = super().get_context_data()
        context['generator'] = self.name
        return context

    def output_to_file(self):
        filename = self.get_output_filename()
        if self.report:
            print('->', filename)

        rendered = self.render_output()
        directory = os.path.dirname(filename)
        if not os.path.isdir(directory):
            os.makedirs(directory)
        with open(filename, 'w', encoding='utf8') as output:
            output.write(rendered)

    def get_output_filename(self):
        destination = '%s%s' % (self.flourish.output_dir, self.current_path)
        if destination.endswith('/'):
            destination = destination + 'index.html'
        if not destination.endswith(self.file_extension):
            destination = destination + '.html'
        return destination


class StaticGenerator(BaseGenerator):
    file_extension = '.html'

    def get_objects(self, tokens):
        self.source_objects = []
        return []


class SourceGenerator(SourcesMixin, PageTemplateMixin, BaseGenerator):
    template_name = 'page.html'
    file_extension = '.html'

    def get_context_data(self):
        context = super().get_context_data()
        context['page'] = self.source_objects[0]
        context.update(self.source_objects[0]._config)
        return context


class IndexGenerator(SourcesMixin, BaseGenerator):
    template_name = 'index.html'
    file_extension = '.html'

    def get_context_data(self):
        context = super().get_context_data()
        context['pages'] = self.source_objects
        return context


class PaginatedIndexGenerator(IndexGenerator):
    per_page = 10

    def generate_path(self, tokens):
        path = self.get_current_path(tokens)
        paginator = self.get_paginated_objects(tokens, path)
        for page in paginator:
            self.current_path = page.path
            self.source_objects = page.object_list
            self.current_page = page
            self.output_to_file()

    def get_paginated_objects(self, tokens, path):
        objects = self.get_objects(tokens)
        per_page = self.get_per_page()
        self.paginator = Paginator(objects, per_page, path)
        return self.paginator

    def get_per_page(self):
        return self.per_page

    def get_context_data(self):
        context = super().get_context_data()
        context['current_page'] = self.current_page
        context['pagination'] = self.paginator
        return context



from collections import defaultdict
from datetime import date, datetime, timezone
from operator import attrgetter
import warnings


class SourceList:
    ARGS = [
        'filters',
        'ordering',
        'slice',
        'future',
    ]

    def __init__(self, sources, **kwargs):
        self.sources = sources
        self.ordering = []
        self.filters = []
        self.slice = None
        self.future = True

        for arg in self.ARGS:
            if arg in kwargs:
                if kwargs[arg] is not None:
                    setattr(self, arg, kwargs[arg])

    def all(self):
        """ Get all source documents. """
        return self

    def count(self):
        return self.__len__()

    def filter(self, **kwargs):
        clone = self.clone()
        for key, value in kwargs.items():
            clone.filters.append((key, value))
        return clone

    def exclude(self, **kwargs):
        clone = self.clone()
        for key, value in kwargs.items():
            if '__' in key:
                field, operator = key.split('__', 2)
            else:
                field, operator = key, 'eq'
            new_operator = INVERSE_OPERATORS.get(operator)
            key = '%s__%s' % (field, new_operator)
            clone.filters.append((key, value))
        return clone

    def exclude_future(self):
        clone = self.clone()
        clone.filters.append(('published__lt', datetime.now(tz=timezone.utc)))
        return clone

    def order_by(self, *args):
        return self.clone(ordering=args)

    def clone(self, **kwargs):
        for arg in self.ARGS:
            value = getattr(self, arg)
            if type(value) == list:
                kwargs.setdefault(arg, value[:])
            else:
                kwargs.setdefault(arg, value)
        return type(self)(self.sources, **kwargs)

    def get_filtered_sources(self):
        sources = []
        for source in self.sources:
            add = True
            if 'published' in source and not self.future:
                if source['published'] > datetime.now(tz=timezone.utc):
                    add = False
            if len(self.filters):
                for filter in self.filters:
                    if not add:
                        # already filtered out by a previous filter
                        continue
                    if '__' in filter[0]:
                        field, operator = filter[0].split('__', 2)
                    else:
                        field, operator = filter[0], 'eq'
                    operation = OPERATORS.get(operator)
                    if operation is None:
                        # FIXME write test and raise more useful error
                        raise RuntimeError
                    is_date_filter = (
                        field in ['year', 'month', 'day'] and
                        'published' in source and
                        type(source['published'] == datetime)
                    )

                    try:
                        if is_date_filter:
                            test = '%02d' % getattr(
                                source['published'], field)
                        else:
                            test = getattr(source, field)
                    except AttributeError:
                        test = None

                    result = operation(test, filter[1])
                    if not result:
                        add = False
            if add:
                sources.append(source)
        return sources

    @property
    def publication_dates(self):
        def recursively_default_dict():
            return defaultdict(recursively_default_dict)

        _captured = recursively_default_dict()
        for _source in self.all():
            try:
                _date = getattr(_source, 'published')
                _captured[_date.year][_date.month][_date.day] = 1
            except AttributeError:
                pass

        _dates = []
        for _year in sorted(_captured):
            _months = []
            for _month in sorted(_captured[_year]):
                _days = []
                for _day in sorted(_captured[_year][_month]):
                    _days.append(date(_year, _month, _day))
                _months.append({
                    'month': date(_year, _month, 1),
                    'days': _days
                })
            _dates.append({'year': date(_year, 1, 1), 'months': _months})
        return _dates

    def __iter__(self):
        sources = self.get_filtered_sources()
        for order in self.ordering:
            if order.startswith('-'):
                rev = True
                attr = order[1:]
            else:
                rev = False
                attr = order
            try:
                sources = sorted(
                    sources,
                    key=attrgetter(attr),
                    reverse=rev
                )
            except AttributeError:
                warnings.warn(
                    'sorting sources by "%s" failed: '
                    'not all sources have that attribute' % order
                )

        if self.slice is not None:
            sources = sources.__getitem__(self.slice)
        return iter(sources)

    def __len__(self):
        _sources = 0
        for source in self:
            _sources += 1
        return _sources

    def __getitem__(self, item):
        if isinstance(item, slice):
            start = item.start
            stop = item.stop
            if start and start < 0:
                start = self.count() + start
            if stop and stop < 0:
                stop = self.count() + stop
            return self.clone(slice=slice(start, stop))
        if item < 0:
            item = self.count() + item
        iterator = iter(self)
        try:
            for i in range(0, item+1):
                obj = next(iterator)
        except StopIteration:
            raise IndexError(item)
        return obj
        return None


def _equal_or_inside(value, test):
    if type(value) == list:
        return test in value
    else:
        return value == test


def _not_equal_or_inside(value, test):
    return not _equal_or_inside(value, test)


def _less_than(value, test):
    try:
        return value < test
    except TypeError:
        return False


def _less_than_or_equal_to(value, test):
    try:
        return value <= test
    except TypeError:
        return False


def _greater_than(value, test):
    try:
        return value > test
    except TypeError:
        return False


def _greater_than_or_equal_to(value, test):
    try:
        return value >= test
    except TypeError:
        return False


def _contains(value, test):
    try:
        if type(value) == list:
            for _v in value:
                if test in _v:
                    return True
            return False
        else:
            return test in value
    except TypeError:
        return False


def _excludes(value, test):
    return not _contains(value, test)


def _inside(value, test):
    return value in test


def _outside(value, test):
    return value not in test


def _set(value, test):
    return value is not None


def _unset(value, test):
    return value is None


OPERATORS = {
    'eq': _equal_or_inside,
    'neq': _not_equal_or_inside,
    'lt': _less_than,
    'lte': _less_than_or_equal_to,
    'gt': _greater_than,
    'gte': _greater_than_or_equal_to,
    'contains': _contains,
    'excludes': _excludes,
    'in': _inside,
    'notin': _outside,
    'set': _set,
    'unset': _unset,
}
INVERSE_OPERATORS = {
    'neq': 'eq',
    'eq': 'neq',
    'lt': 'gte',
    'lte': 'gt',
    'gt': 'lte',
    'gte': 'lt',
    'contains': 'excludes',
    'excludes': 'contains',
    'in': 'notin',
    'notin': 'in',
    'set': 'unset',
    'unset': 'set',
}



import sys
if sys.version_info < (3, 9):
    from importlib_resources import files
else:
    from importlib.resources import files

this = files('flourish.blueprint')
toolbar = this.joinpath('blueprint_toolbar.html').read_text()
template = this.joinpath('blueprint_template.html').read_text()
